
/*
 4-fach Thermometer
 Realisiert mit ATmega88

 Interner RC-Oszillator mit 8MHz



                   PC6 1  28 PC5   (wird intern als volatiles Bit benutzt)
              LCD0 PD0 2  27 PC4   Setzeingang fr Kalibrierung
              LCD1 PD1 3  26 PC3   LM35-4
              LCD2 PD2 4  25 PC2   LM35-3
              LCD3 PD3 5  24 PC1   LM35-2
              LCD4 PD4 6  23 PC0   LM35-1
                   VCC 7  22 GND
                   GND 8  21 AREF
               Q1  PB6 9  20 AVCC
               Q2  PB7 10 19 PB5   SCK   
              LCD5 PD5 11 18 PB4   MISO  LCD R/W
              LCD6 PD6 12 17 PB3   MOSI  LCD_EnableLower
              LCD7 PD7 13 16 PB2   LCD_EnableUpper
            LCD_RS PB0 14 15 PB1   PWM-Out

 
*/
//#define F_CPU 8000000UL  /* CPU clock in Hertz */
//register unsigned char holeWerte asm("r15"); //<- Funktioniert nicht, da Display.c nix davon wei

#include <inttypes.h>
#include <avr/io.h>
#include <stdint.h>
#include <stdio.h>
#include <util/delay.h>
#include <math.h>
#include <avr/eeprom.h>

#include "typedefs.h"
#include "Display.h"

typedef struct{
  U16 magic;    //hex 00FF
  U16 tief[4];
  U16 hoch[4];
  U16 raum[4];
}TeepromMesswerte;


typedef struct{
  U16 messwert[4];
  double a[4],b[4]; //Gradengleichung: y=ax+b
  U8 volatile holeWerte;    //<<- Funktioniert nicht. Wurde durch PB5 ersetzt.
}Tglob;

volatile Tglob  g;
TeepromMesswerte eepromMesswerte;







//-------------------------------------------------------- 
ISR(ADC_vect){
//-------------------------------------------------------- 
static U16 n=0;
static U8 kanal=0;
static U32 summe=0;
static U16 wert[4];

// ADC hat 10 Bit. Fr jedes weitere Bit ist 4-fach Obersampling notwendig
// 12-Bit: 16-fach
// 13-Bit: 64-fach
// 14-Bit: 256-fach
// 15-Bit: 1024-fach

  n++;
  if (n>8192){ // 4^EXTRABITS  (EXTRABITS=5, 4^5=1024)
    n=0;
    wert[kanal] = summe >>8; //>>EXTRABITS  0..32767
    kanal++;
    if (kanal > 3){
      kanal=0;
      if (PORTC & BIT5){
        g.messwert[0] = wert[0];
        g.messwert[1] = wert[1];
        g.messwert[2] = wert[2];
        g.messwert[3] = wert[3];
        PORTC &= ~BIT5;
        g.holeWerte=0;
      }
    }
    ADMUX = kanal;     //  Uref extern
    summe=0;

  }
  if (n>0){
    summe += ADC;
  }
}

//-------------------------------------------------------- 
void leseEEprom(void){
U8 i;
  eeprom_read_block  	( &eepromMesswerte, 0, sizeof(TeepromMesswerte));
  if (eepromMesswerte.magic != 0xff00){ //Noch keine Kalibrierung gespeichert
    for (i=0; i<4; i++){
      eepromMesswerte.tief[i]=256;
      eepromMesswerte.hoch[i]=16384;
      eepromMesswerte.raum[i]=25000;
    }
#if  0 //Nur fr die Fhler von mir
    eepromMesswerte.raum[0]=21924;
    eepromMesswerte.raum[0]=21444;
    eepromMesswerte.raum[0]=21570;
    eepromMesswerte.raum[0]=21296;
#endif    
  }
}
//-------------------------------------------------------- 
void schreibeEEprom(void){
  eepromMesswerte.magic = 0xff00;
  eeprom_write_block  	( &eepromMesswerte, 0, sizeof(TeepromMesswerte));
}
//-------------------------------------------------------- 
void anzeigen(void){
  do{
    sendMessageToLCD();
  }while( lcdGlob.showMessage >0);
}
//-------------------------------------------------------- 
void berechneAB(void){
U8 i;
double dx,dy;
  
  dx= (double)eepromMesswerte.hoch[0] - (double)eepromMesswerte.tief[0];
  for (i=0; i<4; i++){
    dy= (double)eepromMesswerte.hoch[i] - (double)eepromMesswerte.tief[i];
//    g.a[i] = dy/dx; //<- Drehung der Graden wird nicht benutzt. Steigung fest auf 1
    g.a[i] = 1;
//    g.b[i] = eepromMesswerte.hoch[i] - eepromMesswerte.hoch[0] * g.a[i];
    //Kalibrierung des Offset bei Raumtemperatur
    g.b[i] = (double)eepromMesswerte.raum[i] - (double)eepromMesswerte.raum[0];
//    g.b[i] = eepromMesswerte.tief[i] - eepromMesswerte.tief[0] * g.a[i];
  }
#if 0
  clearFullLcd();
  /*
    WICHTIG: Custom Options[Linker]:  -Wl,-u,vfprintf
     auch die libprintf_flt mit einbinden
     siehe http://www.mikrocontroller.net/articles/FAQ#Aktivieren_der_Floating_Point_Version_von_sprintf_beim_WinAVR_mit_AVR-Studio
  */
  snprintf(lcdGlob.message,140,"1dx=%5.0f 2a1=%5.4f b=%5.1f 3a2=%5.4f b=%5.1f 4a3=%5.4f b=%5.1f",
          dx,g.a[1],g.b[1],g.a[2],g.b[2],g.a[3],g.b[3]);
  lcdGlob.showMessage =1;
  anzeigen();
  _delay_ms(5000);
  _delay_ms(3000);
#endif


}
//-------------------------------------------------------- 
void kalibriere(void){
U32 summe[4];
U8 i,j,n;

  n=0;
  lcdGlob.message[n++] = '';  
  lcdGlob.message[n++] = 'o';
  lcdGlob.message[n++] = '';  
  lcdGlob.message[n++] = '*'; 
  lcdGlob.message[n++] = 'c';  
  lcdGlob.message[n++] = 'a'; 
  lcdGlob.message[n++] = 'l';  
  lcdGlob.message[n++] = '.';  
  lcdGlob.message[n++] = ' ';  
  lcdGlob.message[n] = 0;  



  for (i=0; i<4; i++){
    summe[i]=0;
  }
  for (j=0; j<3; j++){
    g.holeWerte = 1;
    PORTC |=BIT5;
    while (PORTC & BIT5){
    };

    if (g.messwert[0] < 19000){ // < 10C
      lcdWrite('L');
    }else
    if (g.messwert[0] > 26000){ // < 10C
      lcdWrite('H');
    }else
    lcdWrite('R');

    for (i=0; i<4; i++){
      summe[i] += g.messwert[i];
    }
  }   
  if (g.messwert[0] < 19000){ // < 10C
    for (i=0; i<4; i++){
      eepromMesswerte.tief[i] = summe[i] / 3;
    }
    schreibeEEprom();
  }else
  if (g.messwert[0] > 26000){ // < 25C
    for (i=0; i<4; i++){
      eepromMesswerte.hoch[i] = summe[i] / 3;
    }
    schreibeEEprom();
  }else{
    for (i=0; i<4; i++){
      eepromMesswerte.raum[i] = summe[i] / 3;
    }
    schreibeEEprom();
  }
  berechneAB();
}

//-------------------------------------------------------- 
U8 wertAnzeigen(U8 n, S16 wert){ //Liefert neues n
U16 val;
  if (wert < -2900/5){
    lcdGlob.message[n++] = ' ';  
    lcdGlob.message[n++] = '-';  
    lcdGlob.message[n++] = '-';  
    lcdGlob.message[n++] = '.';  
    lcdGlob.message[n++] = '-';  
    lcdGlob.message[n++] = '-';  
  }else{
    if (wert < 0){
      val = -wert*5;
      if (val < 1000) lcdGlob.message[n++] = ' ';
      lcdGlob.message[n++] = '-';
    }else{    
      val = wert*5;
      if (val < 1000) lcdGlob.message[n++] = ' ';
      lcdGlob.message[n++] = ' ';
    }
    lcdGlob.message[n] = 48+val/1000;  
    if (48!=lcdGlob.message[n]) n++;
    val %=1000;
    lcdGlob.message[n++] = 48+val/100;  
    val %=100;
    lcdGlob.message[n++] = '.';  
    lcdGlob.message[n++] = 48+val/10;  
    lcdGlob.message[n++] = 48+val%10;  
  }
  return n;
}
//-------------------------------------------------------- 
U8 kalibrierungAnzeigen(U8 n, U16 val){ //Liefert neues n
  lcdGlob.message[n++] = 48+val/10000;  
  val %=10000;
  lcdGlob.message[n++] = 48+val/1000;  
  val %=1000;
  lcdGlob.message[n++] = 48+val/100;  
  val %=100;
  lcdGlob.message[n++] = 48+val/10;  
  lcdGlob.message[n++] = 48+val%10;  
  lcdGlob.message[n++] = ' ';  
return n;
}
//-------------------------------------------------------- 
int main(void){
//-------------------------------------------------------- 
static int x, tic;
static U8 *ram;
static S16 celsius[4];
static S16 celsiusBisher[4];
static S16 celsiusAnzeige[4];
static S16 sollKalibrieren;
U8 n;

  cli();
  DDRB = BIT0 | BIT1 | BIT2 | BIT3 | BIT4;
  DDRC = BIT5;
  DDRD = 0xFF;

  PORTC |=BIT4; //Pullup fr Kalibriertaster ein

  //ADC:
  ADMUX  = 0;     //  Uref extern
  ADCSRA = (1<<ADEN) | (1<<ADSC) | (1<<ADATE) | (1<<ADIF) | (1<<ADIE)     
         | (1<<ADPS1)| (1<<ADPS2); //8MHZ / 64 = 125KHz
  ADCSRB = 0;  
  DIDR0 = BIT0 | BIT1 | BIT2 | BIT3;


  //Timer 1
  TCCR1A = (1<<COM1A0) | (1<<WGM11) | (1<<WGM10) ; //Toggle OC1A, Fast PWM
  TCCR1B = (1<<WGM13) |(1<<WGM12) | (1<<CS11)| (1<<CS10);    //Fast PWM,  Teiler 64 (125KHz)
  OCR1A = 125;  //1KHz
  TIMSK1=0;

   //Globale Variablen initialisieren
//    memset(&g, 0, sizeof(Tglob));               Da rastet der Chip aus
  ram=(unsigned char  *)&g;
  for (x=0; x< sizeof(Tglob); ram[x++]=0);   //Mach ich's halt von Hand

  _delay_ms(300);
  lcdInit();
  leseEEprom();
  berechneAB();
  sei();
  
  clearFullLcd();
  sendStringToLCD("4-fach Thermometer");
  
  home2Lcd();
  sendStringToLCD("D. Schroeder 2010 (V1.4)");

  _delay_ms(2000);

  //Kalibrierung aus EEPROM anzeigen
  clearFullLcd();
  sendStringToLCD("Kalibrierung:");
  home2Lcd();
  n=kalibrierungAnzeigen(0, eepromMesswerte.raum[0]);
  n=kalibrierungAnzeigen(n, eepromMesswerte.raum[1]);
  n=kalibrierungAnzeigen(n, eepromMesswerte.raum[2]);
  n=kalibrierungAnzeigen(n, eepromMesswerte.raum[3]);
  lcdGlob.message[n++] = 0;  
  lcdGlob.showMessage =1;

  anzeigen();

  //9s anzeigen 
  _delay_ms(6000); //(_delay_ms geht aber nur bis ca. 6.5s) 
  _delay_ms(3000);


  clearFullLcd();
  sendStringToLCD("Messe...");
  sollKalibrieren=0;
  

  //-------------------------endlos------------
  while(1){

    if (0==(PINC & BIT4)){
      sollKalibrieren++;
      if (sollKalibrieren>50){  //Taste lange drcken zum kalibrieren
        sollKalibrieren=0;
        kalibriere();  
      }
    }else{
      sollKalibrieren=0;
    }
    _delay_ms(10);
    tic++;
    if (tic>9){
      tic=0;
      g.holeWerte = 1;
  
      PORTC |=BIT5;
      while (PORTC & BIT5){
      };

      #define TMAX  42.4
      #define TMIN (-27.7)

       //0=>-27,7   32768 => +42.4
      for (n=0; n<4; n++){
//        celsius[n] = lrint(100.0* ((((float)g.messwert[n]- g.b[n]) / g.a[n] ) * (TMAX-TMIN) / 32768.0 + TMIN));
        celsius[n] = lrint(100.0* ((((float)g.messwert[n]- g.b[n]) ) * (TMAX-TMIN) / 32768.0 + TMIN));
        if ( abs(celsiusBisher[n] - celsius[n]) > 4){
          celsiusBisher[n] = celsius[n];
          celsiusAnzeige[n] = (celsius[n]+2) / 5;
        }
      }
    
/*Spezialzeichen im String:
  o  = Oberer Display-Controller (Zeile1+2)
  u  = Unterer Display-Controller (Zeile 3+4)
  *  = ClearLCD (auf gerade eingestelltem Controller)
  1  = Home 1. Zeile 
  2  = Home 2. Zeile
  3  = Home 3. Zeile
  4  = Home 4. Zeile
  Gyx= GotoYX x:1..4  y:@..Z
  C  = Cursor Ein (auf gerade eingestelltem Controller) 
  B  = Cursor Ein + Blinken(auf gerade eingestelltem Controller) 
  c  = Cursor Aus (auf gerade eingestelltem Controller) 
  Die HOME-Befehle schalten auch die Displayhlfte um
*/
      n=0;
      lcdGlob.message[n++] = '';  
      lcdGlob.message[n++] = 'o';
      lcdGlob.message[n++] = '';  
      lcdGlob.message[n++] = '*'; 
      lcdGlob.message[n++] = '';  
      lcdGlob.message[n++] = 'u'; 
      lcdGlob.message[n++] = '';  
      lcdGlob.message[n++] = '*'; 

#if 0
      for (i=0; i<4; i++){
        lcdGlob.message[n++] = '';  
        lcdGlob.message[n++] = 49+i; //52-i; //49+i;  

        val = g.messwert[i];
        lcdGlob.message[n++] = 48+val/10000;  
        val %=10000;
        lcdGlob.message[n++] = 48+val/1000;  
        val %=1000;
        lcdGlob.message[n++] = 48+val/100;  
        val %=100;
        lcdGlob.message[n++] = 48+val/10;  
        lcdGlob.message[n++] = 48+val%10;  
        lcdGlob.message[n++] = ' ';  
  
        if (celsius[i] < 0){
          val = -celsius[i];
          lcdGlob.message[n++] = '-';
        }else{    
          val = celsius[i];
          lcdGlob.message[n++] = '+';
        }
        lcdGlob.message[n++] = 48+val/1000;  
        val %=1000;
        lcdGlob.message[n++] = 48+val/100;  
        val %=100;
        lcdGlob.message[n++] = '.';  
        lcdGlob.message[n++] = 48+val/10;  
        lcdGlob.message[n++] = 48+val%10;  
        lcdGlob.message[n++] = '';  
        lcdGlob.message[n++] = 'C';  


        lcdGlob.message[n++] = ' ';  
  
        if (celsiusAnzeige[i] < -2900/5){
          lcdGlob.message[n++] = ' ';  
          lcdGlob.message[n++] = '-';  
          lcdGlob.message[n++] = '-';  
          lcdGlob.message[n++] = '.';  
          lcdGlob.message[n++] = '-';  
          lcdGlob.message[n++] = '-';  
          lcdGlob.message[n++] = '';  
          lcdGlob.message[n++] = 'C';  
        }else{
          if (celsiusAnzeige[i] < 0){
            val = -celsiusAnzeige[i]*5;
            if (val < 1000) lcdGlob.message[n++] = ' ';
            lcdGlob.message[n++] = '-';
          }else{    
            val = celsiusAnzeige[i]*5;
            if (val < 1000) lcdGlob.message[n++] = ' ';
            lcdGlob.message[n++] = ' ';
          }
          lcdGlob.message[n] = 48+val/1000;  
          if (48!=lcdGlob.message[n]) n++;
          val %=1000;
          lcdGlob.message[n++] = 48+val/100;  
          val %=100;
          lcdGlob.message[n++] = '.';  
          lcdGlob.message[n++] = 48+val/10;  
          lcdGlob.message[n++] = 48+val%10;  
          lcdGlob.message[n++] = '';  
          lcdGlob.message[n++] = 'C';  
        }
      }
#endif
 /*

@ABCDEFGHIJKLMNOPQRSTUVWXYZ
123456789 123456789 1234567

      Auen  Innen  Diff
zum WT  12.34  12.34  12.34
vom WT  12.34  12.34  12.34
 12.34  12.34  12.34  32.12

  
 */


        //Gyx= GotoYX x:1..4  y:@..Z
        lcdGlob.message[n++] = '';  
        lcdGlob.message[n++] = 'G';  
        lcdGlob.message[n++] = '8';  
        lcdGlob.message[n++] = 'H';  

        lcdGlob.message[n++] = 'A';  
        lcdGlob.message[n++] = 'u';  
        lcdGlob.message[n++] = '';  
        lcdGlob.message[n++] = 'e';  
        lcdGlob.message[n++] = 'n';  
        lcdGlob.message[n++] = ' ';  
        lcdGlob.message[n++] = 'I';  
        lcdGlob.message[n++] = 'n';  
        lcdGlob.message[n++] = 'n';  
        lcdGlob.message[n++] = 'e';  
        lcdGlob.message[n++] = 'n';  
        lcdGlob.message[n++] = ' ';  

        if (0==(PINC & BIT4)){
          lcdGlob.message[n++] = '*';  //* vor Diff zeigt an, dass Cal. Taste erkannt wurde
        }else{
          lcdGlob.message[n++] = ' ';  
        } 
        lcdGlob.message[n++] = 'D';  
        lcdGlob.message[n++] = 'i';  
        lcdGlob.message[n++] = 'f';  
        lcdGlob.message[n++] = 'f';  

        lcdGlob.message[n++] = '';  
        lcdGlob.message[n++] = '2';  
        lcdGlob.message[n++] = 'z';  
        lcdGlob.message[n++] = 'u';  
        lcdGlob.message[n++] = 'm';  
        lcdGlob.message[n++] = ' ';  
        lcdGlob.message[n++] = 'W';  
        lcdGlob.message[n++] = 'T';  
        lcdGlob.message[n++] = ' ';  
        n=wertAnzeigen(n, celsiusAnzeige[0]);
        n=wertAnzeigen(n, celsiusAnzeige[1]);
        n=wertAnzeigen(n, celsiusAnzeige[1]-celsiusAnzeige[0]);

        lcdGlob.message[n++] = '';  
        lcdGlob.message[n++] = '3';  
        lcdGlob.message[n++] = 'v';  
        lcdGlob.message[n++] = 'o';  
        lcdGlob.message[n++] = 'm';  
        lcdGlob.message[n++] = ' ';  
        lcdGlob.message[n++] = 'W';  
        lcdGlob.message[n++] = 'T';  
        lcdGlob.message[n++] = ' ';  
        n=wertAnzeigen(n, celsiusAnzeige[2]);
        n=wertAnzeigen(n, celsiusAnzeige[3]);
        n=wertAnzeigen(n, celsiusAnzeige[3]-celsiusAnzeige[2]);

        lcdGlob.message[n++] = '';  
        lcdGlob.message[n++] = '4';  
        lcdGlob.message[n++] = ' ';  
        n=wertAnzeigen(n, celsiusAnzeige[2]-celsiusAnzeige[1]);
        n=wertAnzeigen(n, celsiusAnzeige[2]-celsiusAnzeige[0]);
        n=wertAnzeigen(n, celsiusAnzeige[3]-celsiusAnzeige[1]);
        n=wertAnzeigen(n, celsiusAnzeige[3]-celsiusAnzeige[0]);

      lcdGlob.message[n++] = 0;  
      lcdGlob.showMessage =1;
    }
    anzeigen();
   
  }//while
}
//EOF
